"""Collect the errors from contract check results."""

from pathlib import Path
from typing import Iterator

from plshandle._cli import CLIResult


def _is_handled(res, is_strict_mode):
    return res.level == 1 if is_strict_mode else True if res.is_handled else res.is_propagated


def _get_unhandled(results, is_strict_mode):
    return [res for res in results if not _is_handled(res, is_strict_mode)]


def collect_errors(output: CLIResult) -> Iterator[str]:
    """Collect the errors from the contract check results and yield corresponding error messages."""
    for result, report, unhandled in [
        (result, report, unhandled)
        for result in output.results
        for report in result.reports
        for unhandled in _get_unhandled(report.results, output.config.strict)
    ]:
        if unhandled.is_handled and output.config.strict and unhandled.level != 1:
            msg = "{path}:{line}: {exc} not handled at level 1"
        else:
            msg = "{path}:{line}: Violated contract of {func}. Not handled nor propagated {exc}"
        yield msg.format(
            path=Path(result.source.path),
            line=report.context.line,
            func=report.contract.function.fullname,
            exc=unhandled.exception.fullname,
        )



"""Determine the exit code from CLI arguments and check results."""

from plshandle._cli import CLIResult
from plshandle._cli_utils.collect_errors import _get_unhandled


def determine_exit_code(output: CLIResult):
    """Determine the exit code from CLI arguments and check results."""
    if output.config.help_requested:
        return 20
    if output.config.version:
        return 21
    if not output.modules:
        return 10
    if not output.contracts:
        return 11
    if not any(result.reports for result in output.results):
        return 12
    if any(
        _get_unhandled(report.results, output.config.strict)
        for result in output.results
        for report in result.reports
    ):
        return 1
    return 0



"""Create a contract between caller and function that requires the caller to handle specific
exceptions raised by the function.
"""

import sys

from plshandle._cli import cli
from plshandle._cli_utils.build_json import build_json
from plshandle._cli_utils.build_version import build_version
from plshandle._cli_utils.collect_errors import collect_errors
from plshandle._cli_utils.collect_verbose_messages import collect_verbose_messages
from plshandle._cli_utils.determine_exit_code import determine_exit_code
from plshandle._version import __version__


if __name__ == "__main__":
    cli_output = cli(sys.argv[1:])

    if cli_output.config.verbose:
        for msg in collect_verbose_messages(cli_output):
            print(msg)

    if cli_output.config.version:
        print(build_version())
    elif cli_output.config.help_requested:
        pass  # prevent printing the rest if help was requested
    elif not cli_output.modules:
        print("error: No modules found", file=sys.stderr)
    elif not cli_output.contracts:
        print("error: No contracts found", file=sys.stderr)
    elif not any(result.reports for result in cli_output.results):
        print("error: No contracts checked", file=sys.stderr)

    if cli_output.config.json:
        print(build_json(cli_output.results))

    for msg in collect_errors(cli_output):
        print(msg, file=sys.stderr)

    sys.exit(determine_exit_code(cli_output))



"""Check whether all contracts are fulfilled in all modules."""

from dataclasses import dataclass
from typing import Sequence, List

from mypy.modulefinder import BuildSource
from mypy.nodes import Context, FuncDef, TryStmt, Decorator, CallExpr, SymbolNode, TypeInfo

from mypy_extensions import mypyc_attr

from plshandle._cache import MypyCache
from plshandle._visitors.contract_collector import Contract
from plshandle._visitors.alias_resolver import AliasResolver
from plshandle._visitors.scope_tracker import ScopeTracker
from plshandle._ast_utils.resolve_called_functions import resolve_called_functions
from plshandle._ast_utils.resolve_contract import resolve_contract
from plshandle._ast_utils.resolve_handled_types import resolve_handled_types


@dataclass(frozen=True, repr=False)
class ExceptionResult:
    """Determines whether and when the exception is handled."""

    exception: SymbolNode
    is_propagated: bool  #: does the function calling the function propagate this contract?
    is_handled: bool  #: is the exception handled at all? is false if ``is_propagated`` is true
    level: int  #: the level on which the exception is handled (1 = directly above, 0 = not handled)

    def __repr__(self):
        return "ExceptionResult(exception={}, is_propagated={}, is_handled={}, level={})".format(
            self.exception.fullname, self.is_propagated, self.is_handled, self.level
        )


@dataclass(frozen=True, repr=False)
class ContractReport:
    """Report about a contract's fulfillment."""

    contract: Contract
    context: Context
    scope: SymbolNode
    results: Sequence[ExceptionResult]

    def __repr__(self):
        return "ContractReport(contract={}, context={}, scope={}, results={})".format(
            self.contract,
            "Context(line={}, column={}, end_line={})".format(
                self.context.line, self.context.column, self.context.end_line
            ),
            self.scope.fullname,
            self.results,
        )


@dataclass(frozen=True)
class CheckResult:
    """Result of a check for a single module."""

    source: BuildSource
    reports: Sequence[ContractReport]


@dataclass(init=False)
class _CheckerState:
    # ugly per-source state
    module: str
    root: SymbolNode
    reports: List[ContractReport]

    def __init__(self, source: BuildSource, cache: MypyCache):
        self.module = source.module
        self.root = cache.build.files[source.module]
        self.reports = []


@mypyc_attr(allow_interpreted_subclasses=True)
class ContractChecker(ScopeTracker, AliasResolver):
    """Check whether all contracts are fulfilled in all modules."""

    def __init__(
        self, contracts: Sequence[Contract], sources: Sequence[BuildSource], cache: MypyCache
    ):
        super().__init__()
        self.contracts = contracts
        self.cache = cache
        self.results: List[CheckResult] = []

        # traverse all nodes and populate self.results
        for source in sources:
            self.current_state = _CheckerState(source, cache)
            self.visit_mypy_file(self.current_state.root)
            self.results.append(CheckResult(source, self.current_state.reports))

    def visit_call_expr(self, o: CallExpr):
        super().visit_call_expr(o)

        functions = tuple(resolve_called_functions(o, self, self.cache.build.types))
        self.current_state.reports.extend(self._get_reports(o, functions))

    def _is_propagated(self, decorator: Decorator, exception: TypeInfo):
        return any(
            type_ == exception
            for type_ in resolve_contract(
                decorator, self, self.cache.build.types, self.current_state.module
            )
        )

    def _is_handled(self, try_: TryStmt, exception: TypeInfo):
        return any(
            type_ == exception
            for type_ in resolve_handled_types(
                try_, self.cache.build.types, self.current_state.module
            )
        )

    def _check_exception(self, exception: TypeInfo):
        for stmt, level in self.traverse_scope():
            if isinstance(stmt, TryStmt) and self._is_handled(stmt, exception):
                return ExceptionResult(exception, False, True, level)
            if isinstance(stmt, Decorator) and self._is_propagated(stmt, exception):
                return ExceptionResult(exception, True, False, 0)

        return ExceptionResult(exception, False, False, 0)

    def _get_reports(self, context: Context, functions: Sequence[FuncDef]):
        for contract in self.contracts:
            if contract.function in functions:
                yield ContractReport(
                    contract,
                    context,
                    self.determine_current_node(self.current_state.root),
                    [self._check_exception(e) for e in contract.exception_types],
                )



"""Build a JSON array from contract check results."""

import json
from typing import Sequence

from plshandle._cli import CheckResult


def build_json(results: Sequence[CheckResult]) -> str:
    """Build a JSON array from contract check results."""
    return json.dumps(
        [
            {
                "source": {"path": result.source.path, "module": result.source.module,},
                "reports": [
                    {
                        "contract": {
                            "function": report.contract.function.fullname,
                            "exceptions": [exc.fullname for exc in report.contract.exception_types],
                            "source": {
                                "path": report.contract.source.path,
                                "module": report.contract.source.module,
                            },
                        },
                        "context": {
                            "scope": report.scope.fullname,
                            "line": report.context.line,
                            "column": report.context.column,
                        },
                        "results": [
                            {
                                "exception": exc_result.exception.fullname,
                                "is_propagated": exc_result.is_propagated,
                                "is_handled": exc_result.is_handled,
                                "level": exc_result.level,
                            }
                            for exc_result in report.results
                        ],
                    }
                    for report in result.reports
                ],
            }
            for result in results
        ]
    )



"""Collect verbose messages from CLI output."""

import sys
from typing import Iterable, Iterator

from plshandle._cli import CLIResult


def _verbose_list(prefix: str, items: Iterable):
    return "{}:\n- {}\n".format(prefix, "\n- ".join([repr(item) for item in items]) or "<none>")


def collect_verbose_messages(output: CLIResult) -> Iterator[str]:
    """Collect verbose messages from CLI output."""
    yield "CLI args merged with config: {}\n".format(output.config)
    yield _verbose_list("sys.path", sys.path)
    yield _verbose_list("collected modules", output.modules)
    yield _verbose_list("collected contracts", output.contracts)
    yield _verbose_list("contract check results", output.results)



"""Resolve aliases while keeping the scope in mind."""

from functools import reduce
from typing import List, Dict

from mypy.nodes import NameExpr, FuncDef, AssignmentStmt, SymbolNode, RefExpr
from mypy.traverser import TraverserVisitor

from mypy_extensions import mypyc_attr


@mypyc_attr(allow_interpreted_subclasses=True)
class AliasResolver(TraverserVisitor):
    """Resolve aliases while keeping the scope in mind."""

    def __init__(self):
        super().__init__()
        self.alias_scopes: List[Dict[SymbolNode, SymbolNode]] = [{}]

    def visit_func_def(self, o: FuncDef):
        self.alias_scopes.append({})
        super().visit_func_def(o)
        self.alias_scopes.pop()

    def visit_assignment_stmt(self, o: AssignmentStmt):
        super().visit_assignment_stmt(o)

        if (
            len(o.lvalues) == 1
            and isinstance(o.lvalues[0], NameExpr)
            and o.lvalues[0].node is not None
            and isinstance(o.rvalue, RefExpr)
            and o.rvalue.node is not None
        ):
            self.alias_scopes[-1][o.lvalues[0].node] = o.rvalue.node

    def resolve_alias(self, alias: SymbolNode):
        """Resolve the given node alias or returns itself if no alias."""
        all_aliases = reduce(lambda x, y: {**x, **y}, self.alias_scopes)
        while alias in all_aliases:
            alias = all_aliases[alias]
        return alias



"""Decorator used to mark contracts."""

from typing import Type


def _unused(*_, **__):
    pass


def plshandle(*exception_types: Type[BaseException]):
    """Require the caller to handle the given ``exception_types``. This decorator does not modify
    the original function."""
    _unused(exception_types)

    def decorator(function):
        return function

    return decorator



"""Resolve the called function from a call expression."""

from typing import Dict, Optional, Iterator, Tuple, Union

from mypy.nodes import (
    CallExpr,
    Expression,
    RefExpr,
    FuncDef,
    MemberExpr,
    Decorator,
    TypeInfo,
    NameExpr,
)
from mypy.types import (
    Type,
    CallableType,
    Instance,
    Overloaded,
    FunctionLike,
    TypeType,
    UnionType,
)

from plshandle._visitors.alias_resolver import AliasResolver


def _resolve_ref(ref: Expression) -> Optional[Expression]:
    while isinstance(ref, MemberExpr):
        ref = ref.expr
    return ref


def _get_function_from_node(symbol) -> Optional[FuncDef]:
    if isinstance(symbol, Decorator):
        return symbol.func
    if isinstance(symbol, FuncDef):
        return symbol
    return None


def _get_function_from_ref(ref: Expression, resolver: AliasResolver) -> Optional[FuncDef]:
    ref = _resolve_ref(ref)
    if isinstance(ref, RefExpr):
        return _get_function_from_node(resolver.resolve_alias(ref.node))
    return None


def _resolve_unbound_function(callee: Expression, resolver: AliasResolver) -> Optional[FuncDef]:
    return _get_function_from_ref(callee, resolver)


def _try_find_in_type_map(callee: Expression, types: Dict[Expression, Type]) -> Optional[Type]:
    try:
        if isinstance(callee, MemberExpr):
            callee = callee.expr
        return types[callee]
    except KeyError:  # pragma: no cover
        return None


def _type_info_from_expr(expr: Expression) -> Optional[TypeInfo]:
    if isinstance(expr, NameExpr) and isinstance(expr.node, TypeInfo):
        return expr.node
    return None


def _try_resolve_method_call(callee: Expression) -> Optional[Type]:
    # Mypy does not infer the type of expression ``Class()`` correctly if __init__ it is not in
    # mro[0] (maybe that is intended behavior), so manually resolve it. TBD: maybe assist mypy
    # and store it in type map beforehand?

    if isinstance(callee, MemberExpr):
        callee = callee.expr

    if isinstance(callee, CallExpr):
        info = _type_info_from_expr(callee.callee)  # Class()() and Class().method()
        if info:  # pragma: no branch
            return Instance(info, [])
    else:
        info = _type_info_from_expr(callee)
        if info:
            return TypeType(Instance(info, []))  # Class() and Class.method()

    return None


def _find_callee_type(callee: Expression, types: Dict[Expression, Type]) -> Optional[Type]:
    return _try_resolve_method_call(callee) or _try_find_in_type_map(callee, types)


def _resolve_class_types(
    type_: Union[Type, TypeInfo], fallback_method: str = "__init__"
) -> Iterator[Tuple[TypeInfo, Optional[str]]]:
    # mypy states they do not use this type yet, uncomment once they do
    # if isinstance(type_, TypeAliasType):
    #    type_ = type_.expand_all_if_possible()

    if isinstance(type_, TypeInfo):
        yield type_, fallback_method
    else:
        if isinstance(type_, Instance):
            yield from _resolve_class_types(type_.type, "__call__")
        elif isinstance(type_, TypeType):
            yield from _resolve_class_types(type_.item.type, "__init__")
        elif isinstance(type_, UnionType):
            for item in type_.items:
                yield from _resolve_class_types(item, fallback_method)
        elif isinstance(type_, FunctionLike):  # pragma: no branch
            if isinstance(type_, Overloaded):  # pragma: no branch
                type_ = type_.items()[0]
            # just in case Python ever receives another function-like statement
            if isinstance(type_, CallableType):  # pragma: no branch
                yield from _resolve_class_types(type_.ret_type, fallback_method)


def _try_get_callee_method(callee: Expression) -> Optional[str]:
    if isinstance(callee, MemberExpr):
        return callee.name
    return None


def _find_method(class_: TypeInfo, method: str) -> Optional[FuncDef]:
    for base in class_.mro:
        try:
            func = _get_function_from_node(base.names[method].node)
            if func:
                return func
        except (KeyError, AttributeError):
            pass  # o.k, function may be defined in subclass

    return None


def _find_methods(type_: Type, callee: Expression) -> Iterator[FuncDef]:
    for class_, fallback_method in _resolve_class_types(type_):
        method = _find_method(class_, _try_get_callee_method(callee) or fallback_method)
        if method:
            yield method


def _resolve_bound_functions(
    callee: Expression, types: Dict[Expression, Type]
) -> Iterator[FuncDef]:
    yield from _find_methods(_find_callee_type(callee, types), callee)


def resolve_called_functions(
    call: CallExpr, resolver: AliasResolver, types: Dict[Expression, Type]
) -> Iterator[FuncDef]:
    """Yield the called functions from a call expression. This might yield multiple functions
    since the underlying callee might be an union type.
    """

    function = _resolve_unbound_function(call.callee, resolver)
    if function:
        yield function
    else:
        yield from _resolve_bound_functions(call.callee, types)



"""Resolve handled types in a try statement."""

from typing import Dict, Iterator

from mypy.types import Type
from mypy.nodes import Expression, TypeInfo, TryStmt

from plshandle._ast_utils.resolve_exception_types import resolve_exception_types


def resolve_handled_types(
    try_: TryStmt, types: Dict[Expression, Type], module: str
) -> Iterator[TypeInfo]:
    """Yield handled exception types in a try statement."""
    for handler_type, handler_context in zip(try_.types, try_.handlers):
        try:
            yield from resolve_exception_types(types[handler_type], handler_context, module)
        except KeyError:  # pragma: no cover
            pass  # o.k., might be code that does not even concern us



"""Command-line interface of plshandle."""

from argparse import ArgumentParser
from dataclasses import dataclass
import sys
from typing import Optional, List, Sequence, Tuple

from mypy.options import Options

from plshandle._cache import MypyCache
from plshandle._cli_utils.config import read_and_merge_config, Config
from plshandle._visitors.contract_checker import ContractChecker, CheckResult
from plshandle._visitors.contract_collector import ContractCollector, Contract
from plshandle._gather_modules import gather_modules, BuildSource


def _make_arg_parser(description: Optional[str] = None):
    parser = ArgumentParser(description=description or __doc__)
    parser.add_argument(
        "-d",
        "--directory",
        action="append",
        help="additionally collects all modules from all packages in this directory recursively",
    )
    parser.add_argument(
        "-p",
        "--package",
        action="append",
        help="additionally collects all modules from this package recursively",
    )
    parser.add_argument(
        "-m", "--module", action="append", help="additionally include these modules in the check"
    )
    parser.add_argument(
        "--config",
        help="use a config file (CLI args override/extend it) (default: ./pyproject.toml)",
    )
    parser.add_argument(
        "--strict",
        action="store_true",
        # pylint: disable=line-too-long
        help="requires the try block and its handlers to be exactly one level above the function call",
    )
    parser.add_argument(
        "--json",
        action="store_true",
        help="prints a JSON array containing all checked contracts and their results to stdout",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="prints additional information during module processing",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="prints the version of plshandle and your Python interpreter",
    )
    return parser


@dataclass(frozen=True)
class CLIResult:
    """All arguments, collected modules, contracts and check results."""

    config: Config
    modules: Sequence[BuildSource]
    contracts: Sequence[Contract]
    results: Sequence[CheckResult]


def _collect_modules_and_package_roots(args: Config) -> Tuple[Sequence[BuildSource], List[str]]:
    sys.path[:0] = list(args.directory or [])  # be able to find module specs

    package_roots: List[str] = []
    modules = tuple(
        gather_modules(args.directory or [], args.package or [], args.module or [], package_roots)
    )
    return modules, package_roots


def _make_cache(modules: Sequence[BuildSource], package_roots: List[str], options: Options):
    sys.path[:0] = package_roots  # be able to find all modules

    options.package_root = package_roots
    return MypyCache(modules, options)


def cli(args, mypy_options: Options = Options()):
    """Collect all functions decorated with 'plshandle' for all provided
    modules/packages/directories and check whether their callers handle the exceptions passed to
    the decorator. Optionally accepts mypy options.
    """
    try:
        config = read_and_merge_config(_make_arg_parser(cli.__doc__).parse_args(args))
        config.help_requested = False
    except SystemExit:  # pragma: no cover
        config = Config(help_requested=True)
    if config.version or config.help_requested:  # pragma: no cover
        return CLIResult(config, [], [], [])

    modules, package_roots = _collect_modules_and_package_roots(config)
    cache = _make_cache(modules, package_roots, mypy_options)
    contracts = ContractCollector(modules, cache).contracts if modules else []
    results = ContractChecker(contracts, modules, cache).results if contracts else []
    return CLIResult(config, modules, contracts, results)



"""Resolve a contract from a decorator."""

from typing import Dict, Iterator, Optional

from mypy.types import Type
from mypy.nodes import Context, CallExpr, Expression, Decorator, TypeInfo

from plshandle._visitors.alias_resolver import AliasResolver
from plshandle._ast_utils.resolve_called_functions import resolve_called_functions
from plshandle._ast_utils.resolve_exception_types import resolve_exception_types


_PLSHANDLE_QUALIFIER = "plshandle._decorator.plshandle"


def _get_exception_types(
    call: CallExpr, types: Dict[Expression, Type], context: Optional[Context], module: Optional[str]
):
    for arg in call.args:
        try:
            yield from resolve_exception_types(types[arg], context, module)
        except KeyError:  # pragma: no cover
            pass  # o.k., might be code that does not even concern us


def resolve_contract(
    decorator: Decorator, resolver: AliasResolver, types: Dict[Expression, Type], module: str
) -> Iterator[TypeInfo]:
    """Resolve a contract from a decorator. Yields all provided exception types. Yields nothing if
    no contract was defined.
    """
    for call in decorator.decorators:
        if isinstance(call, CallExpr):
            for function in resolve_called_functions(call, resolver, types):
                if function.fullname == _PLSHANDLE_QUALIFIER:
                    yield from _get_exception_types(call, types, decorator, module)



"""Keep track of the scope so it can be traversed upwards."""

from typing import List, Iterator, Tuple

from mypy.nodes import Statement, FuncDef, ClassDef, SymbolNode
from mypy.traverser import TraverserVisitor

from mypy_extensions import mypyc_attr


@mypyc_attr(allow_interpreted_subclasses=True)
class ScopeTracker(TraverserVisitor):
    """Keep track of the scope so it can be traversed upwards."""

    def __init__(self):
        super().__init__()
        self.scope: List[Statement] = []

    def visit_decorator(self, o):
        self.scope.append(o)
        super().visit_decorator(o)
        self.scope.pop()

    def visit_func_def(self, o):
        self.scope.append(o)
        super().visit_func_def(o)
        self.scope.pop()

    def visit_try_stmt(self, o):
        self.scope.append(o)
        super().visit_try_stmt(o)
        self.scope.pop()

    def visit_if_stmt(self, o):
        self.scope.append(o)
        super().visit_if_stmt(o)
        self.scope.pop()

    def visit_for_stmt(self, o):
        self.scope.append(o)
        super().visit_for_stmt(o)
        self.scope.pop()

    def visit_while_stmt(self, o):
        self.scope.append(o)
        super().visit_while_stmt(o)
        self.scope.pop()

    def visit_with_stmt(self, o):
        self.scope.append(o)
        super().visit_with_stmt(o)
        self.scope.pop()

    def traverse_scope(self) -> Iterator[Tuple[Statement, int]]:
        """Traverse the current scope upwards."""
        level = 1
        for scope in reversed(self.scope):
            yield scope, level
            level += 1

    def determine_current_node(self, fallback: SymbolNode) -> SymbolNode:
        """Determine the current node, that is a function or class. The fallback is used if not
        in a function nor in a class definition.
        """
        for stmt, _ in self.traverse_scope():
            if isinstance(stmt, (FuncDef, ClassDef)):
                return stmt

        return fallback



"""Build version string containing plshandle version, interpreter type and Python version."""

import sys

from plshandle._version import __version__


def _version_to_string(version_info):
    return "{}.{}.{}".format(version_info.major, version_info.minor, version_info.micro)


def build_version():
    """Build version string containing plshandle version, interpreter type and Python version."""
    return "plshandle {version}, {intp} {intp_version}, Python {py_version}".format(
        version=__version__,
        intp=sys.implementation.name,
        intp_version=_version_to_string(sys.implementation.version),
        py_version=sys.version,
    )



"""Mypy build cache."""

from typing import Sequence

from mypy.build import build
from mypy.fscache import FileSystemCache
from mypy.options import Options
from mypy.modulefinder import BuildSource


class MypyCache:
    """Cache mypy's AST and type maps."""

    def __init__(self, sources: Sequence[BuildSource], options: Options = Options()):
        # we need these to traverse the AST later on
        options.preserve_asts = True
        options.export_types = True
        options.check_untyped_defs = True

        # make mypy cache type info to improve performance on subsequent runs
        fs_cache = FileSystemCache()
        fs_cache.set_package_root(options.package_root)

        self.build = build(list(sources), options, None, None, fs_cache) if sources else None



"""Gather modules from directories and packages."""

from importlib.util import find_spec
from pathlib import Path
import pkgutil
from typing import Iterable, Iterator, List, Optional

import setuptools

from mypy.modulefinder import BuildSource


def _find_package_dirs(package: str):
    spec = find_spec(package)
    if not spec or not spec.submodule_search_locations:
        raise FileNotFoundError("Cannot find location of package '{}'".format(package))
    return spec.submodule_search_locations


def _gather_modules_in(packages: Iterable[str], package_roots: List[str]) -> Iterator[str]:
    for package in packages:
        dirs = _find_package_dirs(package)
        package_roots.extend(dirs)
        for module_info in pkgutil.iter_modules([Path(x).as_posix() for x in dirs]):
            if not module_info.ispkg:
                yield "{}.{}".format(package, module_info.name)


def _is_valid_origin(origin: Optional[str]):
    return origin and (origin == "builtin" or Path(origin).exists())


def _gather_module_infos(modules: Iterable[str]) -> Iterator[BuildSource]:
    for module in modules:
        spec = find_spec(module)
        if not spec or not _is_valid_origin(spec.origin):  # pragma: no cover
            raise FileNotFoundError("Cannot find location of module '{}'".format(module))
        if spec.origin != "builtin":  # pragma: no branch
            yield BuildSource(spec.origin, spec.name, None, None)


def gather_modules(
    directories: Iterable[str],
    packages: Iterable[str],
    modules: Iterable[str],
    package_roots: List[str],
):
    """Gather modules from directories and packages."""
    all_packages = list(packages)
    for directory in directories:
        native_dir = str(Path(directory))
        all_packages.extend(setuptools.find_namespace_packages(native_dir))

    yield from _gather_module_infos(_gather_modules_in(all_packages, package_roots))
    yield from _gather_module_infos(modules)



"""Attempt to resolve a Type[BaseException] from a mypy.types.Type."""

from typing import Optional, Iterator

from mypy.types import Type, Instance, TypeType, TupleType, CallableType, Overloaded
from mypy.nodes import Context, TypeInfo


def _is_exception_type(type_: TypeInfo):
    return any(sub.fullname == "builtins.BaseException" for sub in type_.mro)  # pragma: no branch


def _raise_invalid_type(type_: Type, context: Optional[Context], module: Optional[str]):
    raise TypeError(
        "{}:{}: Invalid exception type '{}'".format(
            module or "?", context.line if context else "?", type_
        )
    )


def resolve_exception_types(
    type_: Type, context: Optional[Context], module: Optional[str]
) -> Iterator[TypeInfo]:
    """Attempt to resolve a Type[BaseException] or Tuple[Type[BaseException], ...] (nested) from a
    mypy.types.Type. Raises ``TypeError`` if failed to resolve.
    """
    if isinstance(type_, Overloaded):
        type_ = type_.items()[0]
    if isinstance(type_, CallableType):
        type_ = TypeType(type_.ret_type, line=type_.line, column=type_.column)

    if isinstance(type_, TupleType):
        for item in type_.items:  # pragma: no branch
            yield from resolve_exception_types(item, context, module)
    elif isinstance(type_, TypeType) and isinstance(type_.item, Instance):  # pragma: no branch
        if _is_exception_type(type_.item.type):
            yield type_.item.type
        else:
            _raise_invalid_type(type_, context, module)



"""Read config from ./pyproject.toml or --config and merge it with the other args."""

from dataclasses import dataclass
from typing import Optional, Iterable

import toml


@dataclass
class Config:
    """Parsed arguments."""

    config: Optional[str] = None
    directory: Optional[Iterable[str]] = None
    package: Optional[Iterable[str]] = None
    module: Optional[Iterable[str]] = None
    strict: bool = False  #: try block + handlers must be one level above call
    json: bool = False  #: print checked contracts as JSON array to stdout
    verbose: bool = False  #: verbose output
    version: bool = False
    help_requested: bool = False


def _read_list(config: dict, key: dict, file: str):
    try:
        dirs = config[key]
        if isinstance(dirs, list):
            return dirs
        raise TypeError("{}: '{}' must be a list".format(file, key))
    except KeyError:
        return []


def _read_bool(config: dict, key: dict, file: str, default=False):
    try:
        value = config[key]
        if isinstance(value, bool):
            return value
        raise TypeError("{}: '{}' must be a boolean".format(file, key))
    except KeyError:
        return default


def _read_from_file(file: str) -> Config:
    config = toml.load(file)
    try:
        config = config["tool"]["plshandle"]
    except KeyError:
        raise FileNotFoundError("No such section 'tool.plshandle' in {}".format(file))

    return Config(
        directory=_read_list(config, "directories", file),
        package=_read_list(config, "packages", file),
        module=_read_list(config, "modules", file),
        strict=_read_bool(config, "strict", file),
        json=_read_bool(config, "json", file),
        verbose=_read_bool(config, "verbose", file),
    )


def read_and_merge_config(cli_args: Config) -> Config:
    """Read config from ./pyproject.toml or --config and merge it with the other args."""
    try:
        cfg_file = cli_args.config or "pyproject.toml"
        cfg_args = _read_from_file(cfg_file)
    except FileNotFoundError as exc:
        if cli_args.config:
            raise exc  # we explicitly specified a config, so raise if FNF
        cfg_args = Config()
        cfg_file = None

    return Config(
        config=cfg_file,
        directory=(cfg_args.directory or []) + (cli_args.directory or []),
        package=(cfg_args.package or []) + (cli_args.package or []),
        module=(cfg_args.module or []) + (cli_args.module or []),
        strict=cfg_args.strict or cli_args.strict,
        json=cfg_args.json or cli_args.json,
        verbose=cfg_args.verbose or cli_args.verbose,
    )



"""Collect contracts defined in modules."""

from dataclasses import dataclass
from typing import Iterable, List

from mypy.modulefinder import BuildSource
from mypy.nodes import FuncDef, Decorator, TypeInfo

from mypy_extensions import mypyc_attr

from plshandle._cache import MypyCache
from plshandle._visitors.alias_resolver import AliasResolver
from plshandle._ast_utils.resolve_contract import resolve_contract


@dataclass(frozen=True, repr=False)
class Contract:
    """Contract created by ``function``, requires handling ``exception_types``."""

    source: BuildSource
    function: FuncDef
    exception_types: Iterable[TypeInfo]

    def __repr__(self):
        return "_Contract(source={}, function={}, exception_types={})".format(
            self.source,
            self.function.fullname,
            "[{}]".format(", ".join([t.fullname for t in self.exception_types])),
        )


@mypyc_attr(allow_interpreted_subclasses=True)
class ContractCollector(AliasResolver):
    """Collect contracts defined in the given sources."""

    def __init__(self, sources: Iterable[BuildSource], cache: MypyCache):
        super().__init__()
        self.types = cache.build.types
        self.contracts: List[Contract] = []

        # traverse all nodes and populate self.contracts
        for source in sources:
            self.source = source
            self.visit_mypy_file(cache.build.files[source.module])

    def visit_decorator(self, o: Decorator):
        super().visit_decorator(o)
        types = tuple(resolve_contract(o, self, self.types, self.source.module))
        if types:
            self.contracts.append(Contract(self.source, o.func, types))



import setuptools

from plshandle import __version__


with open("README.md", "r") as readme:
    setuptools.setup(
        name="plshandle",
        version=__version__,
        author="v7a",
        long_description=readme.read(),
        long_description_content_type="text/markdown",
        url="https://github.com/v7a/plshandle",
        keywords=["exception", "contract", "error handling"],
        install_requires=["mypy >= 0.750", "setuptools >= 41.0", "toml >= 0.10",],
        package_data={"plshandle": ["py.typed"]},
        packages=setuptools.find_namespace_packages(
            exclude=("plshandle.tests", "plshandle.tests.*", "doc", "doc.*")
        ),
        project_urls={
            "Documentation": "https://plshandle.readthedocs.io",
            "Source": "https://github.com/v7a/plshandle",
            "Tracker": "https://github.com/v7a/plshandle/issues",
        },
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Programming Language :: Python :: 3 :: Only",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Topic :: Software Development :: Quality Assurance",
            "Topic :: Software Development :: Testing",
        ],
    )



"""Create a contract between caller and function that requires the caller to handle specific
exceptions raised by the function.
"""

from ._cli import cli
from ._decorator import plshandle
from ._visitors.contract_collector import Contract
from ._visitors.contract_checker import CheckResult, ContractReport, ExceptionResult
from ._version import __version__
